import inspect
import json
import logging
import sys
import time
import types
from datetime import datetime, timezone
from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, Optional, TYPE_CHECKING

import toolz
import ulid

from . import __version__
from .db import save_invocation_in_sqlite
from .filters.attrs import attrs_filter
from .filters.celery import CeleryFilter
from .filters.core import (
    build_frame_filter,
    exec_filter,
    frozen_filter,
    import_filter,
    library_filter,
)
from .filters.exception import ExceptionFilter
from .filters.huey import HueyFilter
from .filters.kolo import kolo_filter
from .filters.logging import LoggingFilter
from .filters.requests import ApiRequestFilter
from .filters.sql import SQLQueryFilter
from .filters.unittest import UnitTestFilter
from .git import COMMIT_SHA
from .serialize import (
    frame_path,
    get_callsite_data,
    get_content,
    get_request_body,
    monkeypatch_queryset_repr,
    serialize_locals,
    serialize_potential_json,
)


logger = logging.getLogger("kolo")


if TYPE_CHECKING:
    from django.http import HttpRequest, HttpResponse

    from .filters.core import AdvancedFrameFilter, FrameFilter
    from .serialize import UserCodeCallSite


def data_merge(values):
    if all(isinstance(value, list) for value in values):
        return list(chain.from_iterable(values))
    return values[0]


class KoloProfiler:
    """
    Collect runtime information about code to view in VSCode.

    include_frames can be passed to enable profiling of standard library
    or third party code.

    ignore_frames can also be passed to disable profiling of a user's
    own code.

    The list should contain fragments of the path to the relevant files.
    For example, to include profiling for the json module the include_frames
    could look like ["/json/"].

    The list may also contain frame filters. A frame filter is a function
    (or other callable) that takes the same arguments as the profilefunc
    passed to sys.setprofile and returns a boolean representing whether
    to allow or block the frame.

    include_frames takes precedence over ignore_frames. A frame that
    matches an entry in each list will be profiled.
    """

    def __init__(
        self, db_path: Path, *, include_frames=(), ignore_frames=(), config=None
    ) -> None:
        self.db_path = db_path
        self.timestamp = datetime.now(timezone.utc)
        self.invocation_id = f"inv_{ulid.new()}"
        self.frames_of_interest: List[Dict[str, Any]] = []
        self.request: Optional[Dict[str, Any]] = None
        self.response: Optional[Dict[str, Any]] = None
        self.config = config if config is not None else {}
        self.include_frames = list(map(build_frame_filter, include_frames))
        self.ignore_frames = list(map(build_frame_filter, ignore_frames))
        self.default_include_frames: List[AdvancedFrameFilter] = [
            CeleryFilter(self.config),
            HueyFilter(self.config),
            ApiRequestFilter(self.config),
            ExceptionFilter(self.config),
            LoggingFilter(self.config),
            SQLQueryFilter(self.config),
            UnitTestFilter(self.config),
        ]
        self.default_ignore_frames: List[FrameFilter] = [
            library_filter,
            frozen_filter,
            import_filter,
            exec_filter,
            attrs_filter,
            kolo_filter,
        ]
        self.call_frame_ids: List[Dict[str, str]] = []
        self.use_frame_boundaries = self.config.get("use_frame_boundaries", False)

    def initialize_request(self, request: "HttpRequest"):
        self.request = {
            "frame_id": f"frm_{ulid.new()}",
            "scheme": request.scheme,
            "method": request.method,
            "path_info": request.path_info,
            "body": get_request_body(request),
            "headers": dict(request.headers),
            "url_pattern": None,
            "type": "django_request",
        }
        if self.use_frame_boundaries:
            self.frames_of_interest.append(self.request)

    def set_url_pattern(self, request: "HttpRequest"):
        assert isinstance(self.request, dict)
        match = request.resolver_match
        # match can be None before url resolving, but we always
        # call set_url_pattern after resolving
        assert match
        self.request["url_pattern"] = {
            "namespace": match.namespace,
            "route": match.route,
            "url_name": match.url_name,
            "view_path": match._func_path,
        }

    def finalize_response(self, response: "HttpResponse") -> None:
        duration = datetime.now(timezone.utc) - self.timestamp

        self.response = {
            "frame_id": f"frm_{ulid.new()}",
            "ms_duration": round(duration.total_seconds() * 1000, 2),
            "status_code": response.status_code,
            "content": get_content(response),
            "headers": dict(response.items()),
            "type": "django_response",
        }
        if self.use_frame_boundaries:
            self.frames_of_interest.append(self.response)

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> None:
        if event in ["c_call", "c_return"]:
            return

        for frame_filter in self.include_frames:
            try:
                if frame_filter(frame, event, arg):
                    self.process_frame(frame, event, arg)
                    return
            except Exception as e:
                logger.debug(
                    "Unexpected exception in include_frames: %s",
                    frame_filter,
                    exc_info=e,
                )
                continue

        for frame_filter in self.ignore_frames:
            try:
                if frame_filter(frame, event, arg):
                    return
            except Exception as e:
                logger.debug(
                    "Unexpected exception in ignore_frames: %s",
                    frame_filter,
                    exc_info=e,
                )
                continue

        for frame_filter in self.default_include_frames:
            try:
                if frame_filter(frame, event, arg):
                    if frame_filter.use_frames_of_interest:
                        self.process_frame(frame, event, arg)
                    with monkeypatch_queryset_repr():
                        frame_data = frame_filter.process(
                            frame, event, arg, self.call_frame_ids
                        )
                    if frame_data:
                        self.frames_of_interest.append(frame_data)
                    return
            except Exception as e:
                logger.debug(
                    "Unexpected exception in default_include_frames: %s",
                    frame_filter,
                    exc_info=e,
                )
                continue

        for frame_filter in self.default_ignore_frames:
            try:
                if frame_filter(frame, event, arg):
                    return
            except Exception as e:
                logger.debug(
                    "Unexpected exception in default_ignore_frames: %s",
                    frame_filter,
                    exc_info=e,
                )
                continue

        try:
            self.process_frame(frame, event, arg)
        except Exception as e:
            logger.debug(
                "Unexpected exception in KoloProfiler.process_frame",
                exc_info=e,
            )

    def __enter__(self):
        sys.setprofile(self)

    def __exit__(self, *exc):
        sys.setprofile(None)

    def save_request_in_db(self) -> None:
        wal_mode = self.config.get("wal_mode", True)
        json_data = {
            "request_id": self.invocation_id,
            "invocation_id": self.invocation_id,
            "current_commit_sha": COMMIT_SHA,
            "timestamp": str(self.timestamp),
            "frames_of_interest": self.frames_of_interest,
            "meta": {
                "version": __version__,
                "use_frame_boundaries": self.use_frame_boundaries,
            },
        }
        if self.use_frame_boundaries:
            save_invocation_in_sqlite(
                self.db_path, self.invocation_id, json.dumps(json_data), wal_mode
            )
            return

        json_data["request"] = self.request
        json_data["response"] = self.response

        filter_data = (filter.data for filter in self.default_include_frames)
        merged_data = toolz.merge_with(data_merge, json_data, *filter_data)

        try:
            dumped_data = json.dumps(merged_data)
        except TypeError:
            merged_data = json_data
            for frame_filter in self.default_include_frames:
                data = frame_filter.data
                try:
                    json.dumps(data)
                except TypeError:
                    name = type(frame_filter).__name__
                    logger.error(
                        f"Unexpected exception when serializing {name} data: {data}"
                    )
                else:
                    merged_data = toolz.merge_with(data_merge, merged_data, data)

            dumped_data = json.dumps(merged_data)

        save_invocation_in_sqlite(
            self.db_path, self.invocation_id, dumped_data, wal_mode
        )

    def process_frame(self, frame: types.FrameType, event: str, arg: object) -> None:
        user_code_call_site: Optional[UserCodeCallSite]
        if event == "call" and self.call_frame_ids:
            user_code_call_site = get_callsite_data(frame, self.call_frame_ids[-1])
        else:
            # If we are a return frame, we don't bother duplicating
            # information for the call frame.
            # If we are the first call frame, we don't have a callsite.
            user_code_call_site = None

        frame_id = f"frm_{ulid.new()}"
        co_name = frame.f_code.co_name
        if event == "call":
            call_frame_data = {
                "frame_id": frame_id,
                "filepath": frame.f_code.co_filename,
                "co_name": co_name,
            }
            self.call_frame_ids.append(call_frame_data)
        elif event == "return":  # pragma: no branch
            self.call_frame_ids.pop()

        with monkeypatch_queryset_repr():
            self.frames_of_interest.append(
                {
                    "id": frame_id,
                    "path": frame_path(frame),
                    "co_name": co_name,
                    "qualname": get_qualname(frame),
                    "event": event,
                    "arg": serialize_potential_json(arg),
                    "locals": serialize_locals(frame.f_locals),
                    "timestamp": time.time(),
                    "type": "frame",
                    "user_code_call_site": user_code_call_site,
                }
            )


def get_qualname(frame: types.FrameType) -> Optional[str]:
    try:
        qualname = frame.f_code.co_qualname  # type: ignore
    except AttributeError:
        pass
    else:
        module = frame.f_globals["__name__"]
        return f"{module}.{qualname}"

    co_name = frame.f_code.co_name
    if co_name == "<module>":  # pragma: no cover
        module = frame.f_globals["__name__"]
        return f"{module}.<module>"

    try:
        outer_frame = frame.f_back
        assert outer_frame
        try:
            function = outer_frame.f_locals[co_name]
        except KeyError:
            try:
                self = frame.f_locals["self"]
            except KeyError:
                cls = frame.f_locals.get("cls")
                if isinstance(cls, type):
                    function = inspect.getattr_static(cls, co_name)
                else:
                    try:
                        qualname = frame.f_locals["__qualname__"]
                    except KeyError:
                        function = frame.f_globals[co_name]
                    else:  # pragma: no cover
                        module = frame.f_globals["__name__"]
                        return f"{module}.{qualname}"
            else:
                function = inspect.getattr_static(self, co_name)
                if isinstance(function, property):
                    function = function.fget

        return f"{function.__module__}.{function.__qualname__}"
    except Exception:
        return None



import os
import types
from importlib import import_module
from typing import Any, Dict, List, TYPE_CHECKING, Union


if TYPE_CHECKING:
    from typing_extensions import Protocol

    class FrameFilter(Protocol):
        def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
            pass

    ProtoFrameFilter = Union[str, FrameFilter, Dict[str, str]]

    class AdvancedFrameFilter(Protocol):
        config: Dict[str, Any]
        data: Dict[str, Any]
        use_frames_of_interest: bool

        def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
            pass

        def process(
            self,
            frame: types.FrameType,
            event: str,
            arg: object,
            call_frame_ids: List[Dict[str, str]],
        ):
            pass


class HasPath:
    def __init__(self, path: str):
        self.path = path

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return self.path in frame.f_code.co_filename

    def __repr__(self):
        return f'HasPath("{self.path}")'

    def __eq__(self, other):
        return self.path == other.path


def build_frame_filter(filter: "ProtoFrameFilter") -> "FrameFilter":
    if isinstance(filter, str):
        return HasPath(filter)
    if isinstance(filter, dict):
        filter_path = filter["callable"]
        module_path, _sep, filter_name = filter_path.rpartition(".")
        module = import_module(module_path)
        return getattr(module, filter_name)
    return filter


def exec_filter(frame: types.FrameType, event: str, arg: object) -> bool:
    """
    Ignore a frame running a string executed using exec

    We can't show especially interesting information about it, so we skip it.

    A namedtuple is a common example of this case.
    """
    return frame.f_code.co_filename == "<string>"


def frozen_filter(frame: types.FrameType, event: str, arg: object) -> bool:
    """
    Ignore all frozen modules

    Frozen modules are compiled into the interpreter, so are almost
    always part of the standard library.

    To opt into showing a frozen module, specify it in include_frames.
    """
    return "<frozen " in frame.f_code.co_filename


def import_filter(frame: types.FrameType, event: str, arg: object) -> bool:
    """
    Ignore import machinery

    The import system uses frozen modules, which don't have the same
    "lib/python" string fragment in their filepath as the standard
    library or third party code.
    """
    import_modules = (
        "<builtin>/frozen importlib._bootstrap_external",
        "<builtin>/frozen _structseq",
    )
    return frame.f_code.co_filename in import_modules


def library_filter(frame: types.FrameType, *args, **kwargs) -> bool:
    """
    Ignore library code

    We want to not show library calls, so attempt to filter them out here.
    """
    filepath = frame.f_code.co_filename
    paths = (
        "lib/python",
        "lib/pypy",
        "versions/pypy",
        "/PyPy/",
        "/site-packages/",
        "/x64/lib/",
    )
    if any(os.path.normpath(path) in filepath for path in paths):
        return True
    # Something like \python\python310\lib\ on windows
    return (
        os.path.normpath("/Python/") in filepath
        and os.path.normpath("/lib/") in filepath
    )



import types
from typing import Any, Dict, List, Tuple, TYPE_CHECKING

import ulid

from kolo.serialize import serialize_function_args, serialize_function_kwargs


if TYPE_CHECKING:
    # Literal and TypedDict only exist on python 3.8+
    # We run mypy using a high enough version, so this is ok!
    from typing import Literal, TypedDict

    class CeleryJob(TypedDict):
        frame_id: str
        name: str
        args: Tuple[Any, ...]
        kwargs: Dict[str, Any]
        type: Literal["celery_job"]


class CeleryFilter:
    def __init__(self, config) -> None:
        self.config = config
        self.use_frame_boundaries = self.config.get("use_frame_boundaries", False)
        self.use_frames_of_interest = not self.use_frame_boundaries
        self.data: Dict[str, List[CeleryJob]] = {"jobs_enqueued": []}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        filepath = frame.f_code.co_filename
        return (
            "celery" in filepath
            and "sentry_sdk" not in filepath
            and "apply_async" in frame.f_code.co_name
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frame_ids: List[Dict[str, str]],
    ):
        if event == "return":
            return

        frame_locals = frame.f_locals
        task_name = frame_locals["self"].name
        task_args = serialize_function_args(frame_locals["args"])
        task_kwargs = serialize_function_kwargs(frame_locals["kwargs"])

        job: CeleryJob = {
            "frame_id": f"frm_{ulid.new()}",
            "name": task_name,
            "args": task_args,
            "kwargs": task_kwargs,
            "type": "celery_job",
        }

        if self.use_frame_boundaries:
            return job
        else:
            self.data["jobs_enqueued"].append(job)



import subprocess
from typing import Optional


def get_current_commit_sha() -> Optional[str]:
    try:
        git_process = subprocess.run(
            ["git", "rev-parse", "HEAD"],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except OSError:
        return None

    if git_process.stderr:
        return None  # pragma: no cover
    return git_process.stdout.decode("utf-8").strip()


COMMIT_SHA = get_current_commit_sha()



import os
import types


def kolo_filter(frame: types.FrameType, event: str, arg: object) -> bool:
    """Don't profile kolo code"""
    filename = frame.f_code.co_filename
    paths = ("/kolo/middleware", "/kolo/profiler", "/kolo/serialize")
    return any(os.path.normpath(path) in filename for path in paths)



import gzip
import json
import os
import types
from cgi import parse_header
from collections.abc import Iterator
from contextlib import contextmanager
from typing import Any, Dict, Optional, Tuple, TYPE_CHECKING, Union


if TYPE_CHECKING:
    # TypedDict only exists on python 3.8+
    # We run mypy using a high enough version, so this is ok!
    from typing import TypedDict

    from django.http import HttpRequest, HttpResponse, StreamingHttpResponse

    class UserCodeCallSite(TypedDict):
        line_number: int
        call_frame_id: str


@contextmanager
def monkeypatch_queryset_repr():
    try:
        from django.db.models import QuerySet
    except ImportError:  # pragma: no cover
        yield
        return

    old_repr = QuerySet.__repr__

    def new_repr(queryset):
        if queryset._result_cache is None:
            return f"Unevaluated queryset for: {queryset.model}"
        return old_repr(queryset)

    QuerySet.__repr__ = new_repr  # type: ignore
    try:
        yield
    finally:
        QuerySet.__repr__ = old_repr  # type: ignore


def _serialize_local(local: object) -> object:
    if isinstance(local, Iterator):
        return f"Iterator object (type: {type(local)})"

    try:
        json.dumps(local)
    except (TypeError, OverflowError, ValueError):
        try:
            return repr(local)
        except Exception:
            return "SerializationError"
    return local


def serialize_function_args(args: Tuple[Any, ...]) -> Tuple[Any, ...]:
    return tuple([_serialize_local(arg) for arg in args])


def serialize_function_kwargs(kwargs: Dict[str, Any]) -> Dict[str, Any]:
    return {key: _serialize_local(value) for key, value in kwargs.items()}


def serialize_locals(locals: Dict[str, object]) -> str:
    serialized_locals = {key: _serialize_local(value) for key, value in locals.items()}
    return json.dumps(serialized_locals, indent=2)


def serialize_potential_json(arg: object) -> str:
    if arg is None:
        return repr(arg)

    if isinstance(arg, Iterator):
        return f"Iterator object (type: {type(arg)})"

    try:
        return json.dumps(arg)
    except (TypeError, OverflowError, ValueError):
        return repr(arg)


def decode_header_value(bytes_or_str: Union[bytes, str]) -> str:
    """
    Convert a bytes header value to text.

    Valid header values are expected to be ascii in modern times, but
    ISO-8859-1 (latin1) has historically been allowed.

    https://datatracker.ietf.org/doc/html/rfc7230#section-3.2.4
    """
    if isinstance(bytes_or_str, bytes):
        return bytes_or_str.decode("latin1")
    return bytes_or_str


def get_call_frame(
    frame: types.FrameType, filepath: str, co_name: str
) -> Optional[types.FrameType]:
    """Search back in a frame's stack for the triggering user code frame"""

    while frame.f_back is not None:
        frame = frame.f_back
        if frame.f_code.co_filename == filepath and frame.f_code.co_name == co_name:
            return frame
    return None


def get_callsite_data(
    frame: types.FrameType, call_frame_data: Dict[str, str]
) -> Optional["UserCodeCallSite"]:
    """
    Find the parent user code frame and return its frame_id and line number

    We already have the frame_id available in call_frame_data, but we don't
    know the currently active line number, so we search back in the frame
    stack using the filepath and co_name until we find the call frame itself.
    """
    call_frame = get_call_frame(
        frame, call_frame_data["filepath"], call_frame_data["co_name"]
    )
    if call_frame is None:
        return None
    return {
        "call_frame_id": call_frame_data["frame_id"],
        "line_number": call_frame.f_lineno,
    }


def frame_path(frame: types.FrameType) -> str:
    path = frame.f_code.co_filename
    try:
        relative_path = os.path.relpath(path)
    except ValueError:
        relative_path = path
    return f"{relative_path}:{frame.f_lineno}"


def decode_body(body: Any, request_headers: Dict[str, str]) -> Any:
    """Convert a request body into a json-serializable form."""
    try:
        json.dumps(body)
    except TypeError:
        if isinstance(body, bytes):
            content_type = request_headers.get("Content-Type", "")
            charset = parse_header(content_type)[1].get("charset", "utf-8")
            try:
                return body.decode(charset)
            except UnicodeDecodeError:
                return "<Binary request body>"
        return f"Error: Could not parse request body. Type: {type(body)}"
    return body


def get_content(response: Union["HttpResponse", "StreamingHttpResponse"]) -> str:
    if response.streaming:
        return "<Streaming Response>"

    if TYPE_CHECKING:
        assert isinstance(response, HttpResponse)
    content_encoding = response.get("Content-Encoding")
    if content_encoding == "gzip":
        content = gzip.decompress(response.content)
    else:
        content = response.content
    try:
        return content.decode(response.charset)
    except UnicodeDecodeError:
        return f"<Response with invalid charset ({response.charset})>"


def get_request_body(request: "HttpRequest") -> str:
    try:
        return request.body.decode("utf-8")
    except UnicodeDecodeError:
        return "<Binary request body>"



import logging
import os
import types
from typing import Any, Dict, List

import ulid

from ..serialize import serialize_function_kwargs


formatter = logging.Formatter()


class LoggingFilter:
    use_frames_of_interest = False

    def __init__(self, config) -> None:
        self.config = config
        self.data: Dict[str, Any] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        co_name = frame.f_code.co_name
        filename = frame.f_code.co_filename
        return (
            event == "return"
            and co_name == "_log"
            and os.path.normpath("/logging/") in filename
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frame_ids: List[Dict[str, str]],
    ):
        frame_locals = frame.f_locals
        exc_info = frame_locals["exc_info"]
        traceback = None if exc_info is None else formatter.formatException(exc_info)
        extra = frame_locals["extra"]
        extra = None if extra is None else serialize_function_kwargs(extra)
        return {
            "args": frame_locals["args"],
            "extra": extra,
            "frame_id": f"frm_{ulid.new()}",
            "level": logging.getLevelName(frame_locals["level"]),
            "msg": frame_locals["msg"],
            "stack": formatter.formatStack(frame_locals["sinfo"]),
            "traceback": traceback,
            "type": "log_message",
        }



import logging
import os
from pathlib import Path
from typing import Any, MutableMapping

import toml
from cerberus import Validator


logger = logging.getLogger("kolo")


schema = {
    "filters": {
        "type": "dict",
        "schema": {
            "include_frames": {"type": "list"},
            "ignore_frames": {"type": "list"},
            "ignore_request_paths": {"type": "list"},
        },
    },
    "use_frame_boundaries": {"type": "boolean"},
    "wal_mode": {"type": "boolean"},
}
validator = Validator(schema, allow_unknown=False)


def clear_errors(config: MutableMapping[str, Any], errors) -> None:
    for error in errors:
        key = error.document_path[-1]
        if error.info:
            for info in error.info:
                clear_errors(config[key], info)
        else:
            del config[key]


def load_config_from_toml(path: Path) -> MutableMapping[str, Any]:
    try:
        with open(path) as conf:
            config = toml.load(conf)
    except FileNotFoundError:
        return {}
    if not validator.validate(config):
        logger.warning("Kolo config file has errors: %s", validator.errors)
        clear_errors(config, validator._errors)
    return config


def create_kolo_directory() -> Path:
    """
    Create the kolo directory and contents if they do not exist

    Returns the path to the .kolo directory for convenience.
    """
    kolo_directory = (Path(os.environ.get("KOLO_PATH", ".")) / ".kolo").resolve()
    kolo_directory.mkdir(parents=True, exist_ok=True)
    with open(kolo_directory / ".gitignore", "w") as gitignore:
        gitignore.write("db.sqlite3\n")
        gitignore.write("db.sqlite3-shm\n")
        gitignore.write("db.sqlite3-wal\n")
        gitignore.write(".gitignore\n")
    return kolo_directory



import os
import types


def attrs_filter(frame: types.FrameType, event: str, arg: object) -> bool:
    """
    Ignore attrs generated code

    The attrs library constructs an artificial filename for generated
    class methods like __init__ and __hash__.

    It also uses a fake module without a path when creating a class.
    """
    filename = frame.f_code.co_filename
    if filename:
        return filename.startswith("<attrs generated")

    previous = frame.f_back
    if previous is not None and previous.f_code.co_filename == "":
        previous = previous.f_back
    return previous is not None and previous.f_code.co_filename.endswith(
        os.path.normpath("attr/_make.py")
    )



import os
import time
import types
from typing import Dict, List, Optional, TYPE_CHECKING

import ulid

from ..serialize import get_callsite_data


if TYPE_CHECKING:
    # Literal and TypedDict only exist on python 3.8+
    # We run mypy using a high enough version, so this is ok!
    from typing import Literal, TypedDict
    from ..serialize import UserCodeCallSite

    class QueryInfo(TypedDict, total=False):
        query: str
        query_template: str
        user_code_call_site: Optional[UserCodeCallSite]
        call_timestamp: float
        return_timestamp: float

    class QueryStart(TypedDict):
        frame_id: str
        query_template: str
        user_code_call_site: Optional[UserCodeCallSite]
        call_timestamp: float
        type: Literal["start_sql_query"]

    class QueryEnd(TypedDict, total=False):
        frame_id: str
        query: str
        query_template: str
        user_code_call_site: Optional[UserCodeCallSite]
        return_timestamp: float
        type: Literal["end_sql_query"]


class SQLQueryFilter:
    use_frames_of_interest = False

    def __init__(self, config) -> None:
        self.config = config
        self.use_frame_boundaries = self.config.get("use_frame_boundaries", False)
        self.queries_with_call_site: List[QueryInfo] = []
        self.data = {"queries_with_call_site": self.queries_with_call_site}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        co_name = frame.f_code.co_name
        filename = frame.f_code.co_filename
        return (
            co_name == "_execute"
            and os.path.normpath("/django/db/backends/utils.py") in filename
        ) or co_name == "last_executed_query"

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frame_ids: List[Dict[str, str]],
    ):
        co_name = frame.f_code.co_name
        if event == "call" and co_name == "_execute":
            if call_frame_ids:
                user_code_call_site = get_callsite_data(frame, call_frame_ids[-1])
            else:
                user_code_call_site = None

            if self.use_frame_boundaries:
                query_start: QueryStart = {
                    "frame_id": f"frm_{ulid.new()}",
                    "user_code_call_site": user_code_call_site,
                    "call_timestamp": time.time(),
                    "query_template": frame.f_locals["sql"],
                    "type": "start_sql_query",
                }
                return query_start

            query_data: QueryInfo = {
                "user_code_call_site": user_code_call_site,
                "call_timestamp": time.time(),
                "query_template": frame.f_locals["sql"],
            }
            self.queries_with_call_site.append(query_data)

        elif event == "return":
            if self.use_frame_boundaries:
                if co_name == "_execute":
                    if call_frame_ids:
                        user_code_call_site = get_callsite_data(
                            frame, call_frame_ids[-1]
                        )
                    else:
                        user_code_call_site = None

                    query_end: QueryEnd = {
                        "frame_id": f"frm_{ulid.new()}",
                        "user_code_call_site": user_code_call_site,
                        "return_timestamp": time.time(),
                        "query_template": frame.f_locals["sql"],
                        "type": "end_sql_query",
                    }
                    self.last_query = query_end
                    return query_end
                elif co_name == "last_executed_query":  # pragma: no branch
                    from django.db.backends.base.operations import (
                        BaseDatabaseOperations,
                    )

                    if isinstance(  # pragma: no branch
                        frame.f_locals.get("self"), BaseDatabaseOperations
                    ):
                        assert isinstance(arg, str)
                        self.last_query["query"] = arg

            else:
                query_data = self.queries_with_call_site[-1]

                if co_name == "_execute":
                    query_data["return_timestamp"] = time.time()
                elif co_name == "last_executed_query":  # pragma: no branch
                    from django.db.backends.base.operations import (
                        BaseDatabaseOperations,
                    )

                    if isinstance(  # pragma: no branch
                        frame.f_locals.get("self"), BaseDatabaseOperations
                    ):
                        assert isinstance(arg, str)
                        query_data["query"] = arg



import os
from typing import Any, Dict


class UnitTestFilter:
    use_frames_of_interest = False
    event_types = {
        "startTest": "start_test",
        "stopTest": "end_test",
    }

    def __init__(self, config) -> None:
        self.config = config
        self.data: Dict[str, Any] = {}

    def __call__(self, frame, event, arg):
        filepath = frame.f_code.co_filename
        co_name = frame.f_code.co_name
        return (
            event == "call"
            and os.path.normpath("unittest/result.py") in filepath
            and co_name in self.event_types
        )

    def process(self, frame, event, arg, call_frame_ids):  # pragma: no cover
        co_name = frame.f_code.co_name
        testcase = frame.f_locals["test"]
        return {
            "type": self.event_types[co_name],
            "test_name": testcase._testMethodName,
            "test_class": testcase.__class__.__qualname__,
        }



import json
import sqlite3
from contextlib import contextmanager
from pathlib import Path
from typing import Any, MutableMapping, Optional

import toolz

from .config import create_kolo_directory


@contextmanager
def db_cursor(db_path, wal_mode=True):
    """
    Wrap sqlite's cursor for use as a context manager

    Commits all changes if no exception is raised.
    Always closes the cursor/connection after the context manager exits.
    """
    if wal_mode:
        connection = sqlite3.connect(str(db_path), isolation_level=None)
        connection.execute("pragma journal_mode=wal")
    else:
        connection = sqlite3.connect(str(db_path))  # pragma: no cover
    cursor = connection.cursor()
    try:
        yield cursor
        connection.commit()
    finally:
        cursor.close()
        connection.close()


def get_db_path() -> Path:
    return create_kolo_directory() / "db.sqlite3"


def create_invocations_table(cursor) -> None:
    create_table_query = """
    CREATE TABLE IF NOT EXISTS invocations (
        id text PRIMARY KEY NOT NULL,
        created_at TEXT DEFAULT (STRFTIME('%Y-%m-%d %H:%M:%f', 'NOW')) NOT NULL,
        data text NOT NULL
    );
    """
    create_timestamp_index_query = """
        CREATE INDEX IF NOT EXISTS
        idx_invocations_created_at
        ON invocations (created_at);
        """

    cursor.execute(create_table_query)
    cursor.execute(create_timestamp_index_query)


def create_config_table(cursor) -> None:
    create_config_query = """
    CREATE TABLE IF NOT EXISTS config (
        id INTEGER PRIMARY KEY,
        source text NOT NULL UNIQUE,
        last_updated_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP NOT NULL,
        data text NOT NULL
    );
    """
    cursor.execute(create_config_query)


def save_config_to_db(cursor, config) -> None:
    save_config_query = "INSERT OR REPLACE INTO config(source, data) VALUES(?,?);"
    cursor.execute(save_config_query, ("python", json.dumps(config)))


def merge_or_last(args):
    if all(isinstance(arg, dict) for arg in args):
        return toolz.merge(*args)
    return args[-1]


def load_config_from_db(
    db_path: Path, toml_config: Optional[MutableMapping[str, Any]] = None
) -> MutableMapping[str, Any]:
    """
    Load the kolo config from sqlite

    We return the config defined by VSCode, if available, and fallback
    to the config we saved to the db from `.kolo/conf.toml`.
    """
    if toml_config is None:
        toml_config = {}

    wal_mode = toml_config.get("wal_mode", True)

    load_config_query = "SELECT source, data from config;"
    try:
        with db_cursor(db_path, wal_mode) as cursor:
            cursor.execute(load_config_query)
            rows = cursor.fetchall()
    except sqlite3.OperationalError as e:
        if e.args[0] != "no such table: config":
            raise  # pragma: no cover
        setup_db(toml_config)
        return toml_config

    data = {key: json.loads(value) for key, value in rows}
    python_config = data["python"]
    vscode_config = data.get("vscode", {})
    return toolz.merge_with(merge_or_last, python_config, vscode_config)


def setup_db(config: Optional[MutableMapping[str, Any]] = None) -> Path:
    if config is None:
        config = {}
    db_path = get_db_path()

    wal_mode = config.get("wal_mode", True)

    with db_cursor(db_path, wal_mode) as cursor:
        create_invocations_table(cursor)
        create_config_table(cursor)
        save_config_to_db(cursor, config)

    return db_path


def save_invocation_in_sqlite(
    db_path: Path, invocation_id: str, json_string: str, wal_mode: bool = True
) -> None:
    insert_sql = """
        INSERT OR IGNORE INTO invocations(id, data)
        VALUES(?,?)
        """

    # We can't reuse a connection
    # because we're in a new thread
    with db_cursor(db_path, wal_mode) as cursor:
        cursor.execute(insert_sql, (invocation_id, json_string))



import logging
import os
import types
from datetime import datetime, timezone
from typing import Dict, List, Optional, TYPE_CHECKING

import ulid

from ..serialize import decode_body, decode_header_value


logger = logging.getLogger("kolo")


if TYPE_CHECKING:
    # Literal and TypedDict only exist on python 3.8+
    # We run mypy using a high enough version, so this is ok!
    from typing import Literal, TypedDict

    class ApiRequest(TypedDict):
        frame_id: str
        method: str
        url: str
        method_and_full_url: str
        body: Optional[str]
        headers: Dict[str, str]
        timestamp: str
        type: Literal["outbound_http_request"]

    class BaseApiResponse(TypedDict):
        frame_id: str
        timestamp: str
        status_code: int
        headers: Dict[str, str]
        type: Literal["outbound_http_response"]
        method: str
        url: str
        method_and_full_url: str

    class ApiResponse(BaseApiResponse, total=False):
        body: str

    class BaseApiInfo(TypedDict):
        request: ApiRequest

    class ApiInfo(BaseApiInfo, total=False):
        response: ApiResponse


def get_timestamp() -> str:
    return datetime.now(timezone.utc).isoformat()


def get_full_url(frame_locals) -> str:
    scheme = frame_locals["self"].scheme
    host = frame_locals["self"].host
    url = frame_locals["url"]
    return f"{scheme}://{host}{url}"


class ApiRequestFilter:
    use_frames_of_interest = True

    def __init__(self, config) -> None:
        self.config = config
        self.use_frame_boundaries = self.config.get("use_frame_boundaries", False)
        self.data: Dict[str, List[ApiInfo]] = {"api_requests_made": []}
        self.last_response: Optional[ApiResponse] = None

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        # This is equivalent to:
        # self.match_request(frame) or self.match_response(frame) or self.match_urllib(frame)
        # but avoids the function call overhead, which is important for
        # the performance of Kolo's profiling loop.

        filepath = frame.f_code.co_filename
        callable_name = frame.f_code.co_name
        return (
            (
                callable_name == "urlopen"
                and os.path.normpath("urllib3/connectionpool") in filepath
            )
            or (
                callable_name == "request"
                and os.path.normpath("requests/sessions") in filepath
            )
            or (
                callable_name == "do_open"
                and os.path.normpath("urllib/request") in filepath
            )
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frame_ids: List[Dict[str, str]],
    ):
        if event == "call" and self.match_request(frame):
            return self.process_api_request_made(frame)
        elif event == "return" and self.match_response(frame):
            return self.process_api_response(frame)
        elif self.match_urllib(frame):
            return self.process_urllib(frame, event)

    def match_request(self, frame: types.FrameType) -> bool:
        filepath = frame.f_code.co_filename
        callable_name = frame.f_code.co_name
        return (
            callable_name == "urlopen"
            and os.path.normpath("urllib3/connectionpool") in filepath
        )

    def match_response(self, frame: types.FrameType) -> bool:
        filepath = frame.f_code.co_filename
        callable_name = frame.f_code.co_name
        return (
            callable_name == "urlopen"
            and os.path.normpath("urllib3/connectionpool") in filepath
        ) or (
            callable_name == "request"
            and os.path.normpath("requests/sessions") in filepath
        )

    def match_urllib(self, frame: types.FrameType) -> bool:
        filepath = frame.f_code.co_filename
        callable_name = frame.f_code.co_name
        return (
            callable_name == "do_open"
            and os.path.normpath("urllib/request") in filepath
        )

    def process_api_request_made(self, frame: types.FrameType):
        frame_locals = frame.f_locals

        request_headers = {
            key: decode_header_value(value)
            for key, value in frame_locals["headers"].items()
        }

        full_url = get_full_url(frame_locals)
        method = frame_locals["method"].upper()
        method_and_full_url = f"{method} {full_url}"

        api_request: ApiRequest = {
            "frame_id": f"frm_{ulid.new()}",
            "method": method,
            "url": full_url,
            "method_and_full_url": method_and_full_url,
            "body": decode_body(frame_locals["body"], request_headers),
            "headers": request_headers,
            "timestamp": get_timestamp(),
            "type": "outbound_http_request",
        }

        if self.use_frame_boundaries:
            return api_request
        else:
            api_info: ApiInfo = {"request": api_request}
            self.data["api_requests_made"].append(api_info)

    def process_api_response(self, frame: types.FrameType):
        if frame.f_code.co_name == "urlopen":
            frame_locals = frame.f_locals

            full_url = get_full_url(frame_locals)
            method = frame_locals["method"].upper()
            method_and_full_url = f"{method} {full_url}"

            response = frame_locals["response"]
            api_response: ApiResponse = {
                "frame_id": f"frm_{ulid.new()}",
                "headers": dict(response.headers),
                "method": method,
                "method_and_full_url": method_and_full_url,
                "status_code": response.status,
                "timestamp": get_timestamp(),
                "type": "outbound_http_response",
                "url": full_url,
            }
            self.last_response = api_response
            if self.use_frame_boundaries:
                return api_response
            else:
                self.data["api_requests_made"][-1]["response"] = api_response
        else:
            response = frame.f_locals["resp"]
            assert self.last_response is not None
            self.last_response["body"] = response.text

    def process_urllib(self, frame: types.FrameType, event: str):
        request = frame.f_locals["req"]
        full_url = request.full_url
        method = request.get_method()
        method_and_full_url = f"{method} {full_url}"
        if event == "call":
            request_headers = {
                key: decode_header_value(value) for key, value in request.header_items()
            }

            api_request: ApiRequest = {
                "frame_id": f"frm_{ulid.new()}",
                "method": method,
                "url": full_url,
                "method_and_full_url": method_and_full_url,
                "body": decode_body(request.data, request_headers),
                "headers": request_headers,
                "timestamp": get_timestamp(),
                "type": "outbound_http_request",
            }

            if self.use_frame_boundaries:
                return api_request
            else:
                api_info: ApiInfo = {"request": api_request}
                self.data["api_requests_made"].append(api_info)

        elif event == "return":  # pragma: no branch
            response = frame.f_locals["r"]
            api_response: ApiResponse = {
                "frame_id": f"frm_{ulid.new()}",
                "method": method,
                "url": full_url,
                "method_and_full_url": method_and_full_url,
                "timestamp": get_timestamp(),
                "status_code": response.status,
                "headers": dict(response.headers),
                "type": "outbound_http_response",
            }
            if self.use_frame_boundaries:
                return api_response
            else:
                self.data["api_requests_made"][-1]["response"] = api_response



import os
from typing import Any, Dict, List, Tuple, TYPE_CHECKING

import ulid

from kolo.serialize import serialize_function_args, serialize_function_kwargs


if TYPE_CHECKING:
    # Literal and TypedDict only exist on python 3.8+
    # We run mypy using a high enough version, so this is ok!
    from typing import Literal, TypedDict

    class HueyJob(TypedDict):
        frame_id: str
        name: str
        args: Tuple[Any, ...]
        kwargs: Dict[str, Any]
        type: Literal["huey_job"]


class HueyFilter:
    use_frames_of_interest = False

    def __init__(self, config) -> None:
        self.config = config
        self.use_frame_boundaries = self.config.get("use_frame_boundaries", False)
        self.data: Dict[str, List[HueyJob]] = {"jobs_enqueued": []}

    def __call__(self, frame, event, arg):
        filepath = frame.f_code.co_filename
        co_name = frame.f_code.co_name
        if (
            event == "call"
            and co_name == "__init__"
            and os.path.normpath("/huey/api.py") in filepath
        ):
            from huey.api import Task

            return isinstance(frame.f_locals["self"], Task)
        return False

    def process(self, frame, event, arg, call_frame_ids):
        frame_locals = frame.f_locals
        task_object = frame_locals["self"]
        task_args = serialize_function_args(frame_locals["args"])
        task_kwargs = serialize_function_kwargs(frame_locals["kwargs"])

        job: HueyJob = {
            "frame_id": f"frm_{ulid.new()}",
            "name": f"{task_object.__module__}.{task_object.name}",
            "args": task_args,
            "kwargs": task_kwargs,
            "type": "huey_job",
        }
        if self.use_frame_boundaries:
            return job
        else:
            self.data["jobs_enqueued"].append(job)



import logging
import os
import sys
import threading
from typing import Any, Callable, Dict, Tuple

from django.conf import settings
from django.core.exceptions import MiddlewareNotUsed
from django.http import HttpRequest, HttpResponse

from .config import create_kolo_directory, load_config_from_toml
from .db import load_config_from_db, setup_db
from .profiler import KoloProfiler

logger = logging.getLogger("kolo")

DjangoView = Callable[[HttpRequest], HttpResponse]


class KoloMiddleware:
    def __init__(self, get_response: DjangoView) -> None:
        self.get_response = get_response
        if self.should_enable():
            kolo_directory = create_kolo_directory()
            self.config = load_config_from_toml(kolo_directory / "config.toml")
            self.db_path = setup_db(self.config)
        else:
            message = "Kolo has been disabled via DEBUG setting, KOLO_DISABLE environment variable, or third party profiler"
            raise MiddlewareNotUsed(message)

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if self.check_for_third_party_profiler():
            return self.get_response(request)

        self.config = load_config_from_db(self.db_path, self.config)

        filter_config = self.config.get("filters", {})
        ignore_request_paths = filter_config.get("ignore_request_paths", [])
        for path in ignore_request_paths:
            if path in request.path:
                return self.get_response(request)

        self.profiler = KoloProfiler(
            self.db_path,
            include_frames=filter_config.get("include_frames", []),
            ignore_frames=filter_config.get("ignore_frames", []),
            config=self.config,
        )
        self.profiler.initialize_request(request)

        with self.profiler:
            response = self.get_response(request)

        self.profiler.finalize_response(response)

        threading.Thread(
            target=self.profiler.save_request_in_db, name="kolo-save_request_in_db"
        ).start()

        # eventual todo: restore previously set profile
        return response

    def process_view(
        self,
        request: HttpRequest,
        view_func: DjangoView,
        view_args: Tuple[Any],
        view_kwargs: Dict[str, Any],
    ):
        self.profiler.set_url_pattern(request)

    def check_for_third_party_profiler(self) -> bool:
        profiler = sys.getprofile()
        if profiler:
            logger.warning("Profiler %s is active, disabling KoloMiddleware", profiler)
            return True
        return False

    def should_enable(self) -> bool:
        if settings.DEBUG is False:
            logger.debug("DEBUG mode is off, disabling KoloMiddleware")
            return False

        if os.environ.get("KOLO_DISABLE", "false").lower() not in ["false", "0"]:
            logger.debug("KOLO_DISABLE is set, disabling KoloMiddleware")
            return False

        if self.check_for_third_party_profiler():
            return False

        return True



import traceback
import types
from itertools import starmap
from typing import Dict, List, TYPE_CHECKING

import ulid

try:
    from django.db.models import Model as DjangoModel
except ImportError:

    class DjangoModel:  # type: ignore
        """Stub type so isinstance returns False"""


from ..serialize import frame_path, serialize_locals
from .core import library_filter


if TYPE_CHECKING:
    # Literal and TypedDict only exist on python 3.8+
    # We run mypy using a high enough version, so this is ok!
    from typing import Literal, TypedDict

    class ExceptionFrameInfo(TypedDict):
        path: str
        co_name: str
        locals: str
        expanded_locals: Dict[str, str]

    class RecordedException(TypedDict):
        # Usually contains one string. Last string in the list is always
        # the one indicating which exception occurred
        exception_summary: List[str]
        exception_with_traceback: List[str]
        exception_frames: List[ExceptionFrameInfo]
        frame_id: str
        bottom_exception_frame: ExceptionFrameInfo
        type: Literal["exception"]


class ExceptionFilter:
    use_frames_of_interest = False

    def __init__(self, config) -> None:
        self.config = config
        self.use_frame_boundaries = self.config.get("use_frame_boundaries", False)
        self.data: Dict[str, RecordedException] = {}

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        return (
            event == "call"
            and "django" in frame.f_code.co_filename
            and "handle_uncaught_exception" == frame.f_code.co_name
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frame_ids: List[Dict[str, str]],
    ):
        frame_locals = frame.f_locals
        exc_type, exc_value, exc_traceback = frame_locals["exc_info"]

        recorded_exception_frames = []
        expanded_locals_for_frames = []
        for frame_and_line in traceback.walk_tb(exc_traceback):
            frame = frame_and_line[0]

            if not library_filter(frame):
                frame_locals = frame.f_locals

                expanded_locals = {}
                for key, value in frame_locals.items():
                    if hasattr(value, "__dict__") and isinstance(value, DjangoModel):
                        expanded_locals[key] = vars(value)

                recorded_exception_frames.append(frame)
                expanded_locals_for_frames.append(expanded_locals)

        def serialize_exception_frame(frame, expanded_locals) -> "ExceptionFrameInfo":
            return {
                "path": frame_path(frame),
                "co_name": frame.f_code.co_name,
                "locals": serialize_locals(frame.f_locals),
                "expanded_locals": {
                    key: serialize_locals(value)
                    for key, value in expanded_locals.items()
                },
            }

        exception_with_traceback = traceback.format_exception(
            exc_type, exc_value, exc_traceback
        )

        zipped_frames = zip(recorded_exception_frames, expanded_locals_for_frames)
        exception_frames = list(starmap(serialize_exception_frame, zipped_frames))

        exception: RecordedException = {
            "frame_id": f"frm_{ulid.new()}",
            "exception_summary": traceback.format_exception_only(exc_type, exc_value),
            "exception_with_traceback": exception_with_traceback,
            "exception_frames": exception_frames,
            "bottom_exception_frame": exception_frames[-1],
            "type": "exception",
        }
        if self.use_frame_boundaries:
            return exception
        else:
            self.data["exception"] = exception



import sys
from runpy import run_module, run_path

import click

from .config import create_kolo_directory, load_config_from_toml
from .db import load_config_from_db, setup_db
from .profiler import KoloProfiler


def get_profiler() -> KoloProfiler:
    kolo_directory = create_kolo_directory()
    config = load_config_from_toml(kolo_directory / "config.toml")
    db_path = setup_db(config)
    db_config = load_config_from_db(db_path)
    filter_config = db_config.get("filters", {})

    return KoloProfiler(
        db_path,
        include_frames=filter_config.get("include_frames", []),
        ignore_frames=filter_config.get("ignore_frames", []),
        config=db_config,
    )


def profile_module(profiler: KoloProfiler, module_name: str):
    with profiler:
        run_module(module_name, run_name="__main__", alter_sys=True)


def profile_path(profiler: KoloProfiler, path: str):
    with profiler:
        run_path(path, run_name="__main__")


@click.group()
def cli():
    """Base for all kolo command line commands"""

    # Ensure the current working directory is on the path.
    # Important when running the `kolo` script installed by setuptools.
    # Not really necessary when using `python -m kolo`, but doesn't hurt.
    # Note, we use 1, not 0: https://stackoverflow.com/q/10095037
    # This probably doesn't matter for our use case, but it doesn't hurt.
    sys.path.insert(1, ".")


@cli.command(context_settings={"ignore_unknown_options": True})
@click.argument("path")
@click.argument("args", nargs=-1, type=click.UNPROCESSED)
def run(path, args):
    """
    Profile python code using kolo.

    PATH is the path to the python file or module being profiled.
    """
    if path == "python":
        path, *args = args
        if path == "-m":
            path, *args = args
            module = True
        else:
            module = False
    elif path.endswith(".py"):
        module = False
    else:
        module = True

    # Monkeypatch sys.argv, so the profiled code doesn't get confused
    # Without this, the profiled code would see extra args it doesn't
    # know how to handle.
    sys.argv = [path, *args]

    profiler = get_profiler()
    try:
        if module:
            profile_module(profiler, path)
        else:
            profile_path(profiler, path)
    finally:
        profiler.save_request_in_db()


@cli.command()
def config():
    """Display kolo's configuration."""
    from pprint import pformat

    kolo_directory = create_kolo_directory()
    config_file = kolo_directory / "config.toml"
    config = load_config_from_toml(config_file)
    db_path = setup_db(config)
    db_config = load_config_from_db(db_path)

    click.echo("Kolo config from VSCode (takes precedence)")
    click.echo(pformat(db_config))
    click.echo(f"Kolo config from {config_file}")
    click.echo(pformat(config))


if __name__ == "__main__":
    cli()  # pragma: no cover
